@@ -1,4 +1,4 @@
-/* Copyright (C) 2001, 2002, 2004 Red Hat, Inc.
+/* Copyright (C) 2001, 2002, 2004, 2009 Red Hat, Inc.
    Written by Jakub Jelinek <jakub@redhat.com>, 2001.
 
    This program is free software; you can redistribute it and/or modify
@@ -40,12 +40,12 @@ arm_adjust_dyn (DSO *dso, int n, GElf_Dyn *dyn, GElf_Addr start,
       if (sec == -1)
 	return 0;
 
-      data = read_ule32 (dso, dyn->d_un.d_ptr);
+      data = read_une32 (dso, dyn->d_un.d_ptr);
       /* If .got.plt[0] points to _DYNAMIC, it needs to be adjusted.  */
       if (data == dso->shdr[n].sh_addr && data >= start)
-	write_le32 (dso, dyn->d_un.d_ptr, data + adjust);
+	write_ne32 (dso, dyn->d_un.d_ptr, data + adjust);
 
-      data = read_ule32 (dso, dyn->d_un.d_ptr + 4);
+      data = read_une32 (dso, dyn->d_un.d_ptr + 4);
       /* If .got.plt[1] points to .plt, it needs to be adjusted.  */
       if (data && data >= start)
 	{
@@ -57,7 +57,7 @@ arm_adjust_dyn (DSO *dso, int n, GElf_Dyn *dyn, GElf_Addr start,
 		&& strcmp (strptr (dso, dso->ehdr.e_shstrndx,
 				   dso->shdr[i].sh_name), ".plt") == 0)
 	      {
-		write_le32 (dso, dyn->d_un.d_ptr + 4, data + adjust);
+		write_ne32 (dso, dyn->d_un.d_ptr + 4, data + adjust);
 		break;
 	      }
 	}
@@ -74,9 +74,9 @@ arm_adjust_rel (DSO *dso, GElf_Rel *rel, GElf_Addr start,
     {
     case R_ARM_RELATIVE:
     case R_ARM_JUMP_SLOT:
-      data = read_ule32 (dso, rel->r_offset);
+      data = read_une32 (dso, rel->r_offset);
       if (data >= start)
-	write_le32 (dso, rel->r_offset, data + adjust);
+	write_ne32 (dso, rel->r_offset, data + adjust);
       break;
     }
   return 0;
@@ -96,13 +96,13 @@ arm_adjust_rela (DSO *dso, GElf_Rela *rela, GElf_Addr start,
 	  rela->r_addend += (Elf32_Sword) adjust;
 	  /* Write it to the memory location as well.
 	     Not necessary, but we can do it.  */
-	  write_le32 (dso, rela->r_offset, rela->r_addend);
+	  write_ne32 (dso, rela->r_offset, rela->r_addend);
 	}
       break;
     case R_ARM_JUMP_SLOT:
-      data = read_ule32 (dso, rela->r_offset);
+      data = read_une32 (dso, rela->r_offset);
       if (data >= start)
-	write_le32 (dso, rela->r_offset, data + adjust);
+	write_ne32 (dso, rela->r_offset, data + adjust);
       break;
       break;
     }
@@ -126,18 +126,18 @@ arm_prelink_rel (struct prelink_info *info, GElf_Rel *rel, GElf_Addr reladdr)
     {
     case R_ARM_GLOB_DAT:
     case R_ARM_JUMP_SLOT:
-      write_le32 (dso, rel->r_offset, value);
+      write_ne32 (dso, rel->r_offset, value);
       break;
     case R_ARM_ABS32:
       {
-	if (read_ule32 (dso, rel->r_offset))
+	if (read_une32 (dso, rel->r_offset))
 	  {
 	    error (0, 0, "%s: R_ARM_ABS32 relocs with non-zero addend should not be present in prelinked REL sections",
 		   dso->filename);
 	    return 1;
 	  }
 	rel->r_info = GELF_R_INFO (GELF_R_SYM (rel->r_info), R_ARM_GLOB_DAT);
-	write_le32 (dso, rel->r_offset, value);
+	write_ne32 (dso, rel->r_offset, value);
 	/* Tell prelink_rel routine *rel has changed.  */
 	return 2;
       }
@@ -151,6 +151,25 @@ arm_prelink_rel (struct prelink_info *info, GElf_Rel *rel, GElf_Addr reladdr)
 	return 0;
       error (0, 0, "%s: R_ARM_COPY reloc in shared library?", dso->filename);
       return 1;
+    case R_ARM_TLS_DTPOFF32:
+      write_ne32 (dso, rel->r_offset, value);
+      break;
+    /* DTPMOD32 and TPOFF32 is impossible to predict in shared libraries
+       unless prelink sets the rules.  */
+    case R_ARM_TLS_DTPMOD32:
+      if (dso->ehdr.e_type == ET_EXEC)
+        {
+          error (0, 0, "%s: R_ARM_TLS_DTPMOD32 reloc in executable?",
+                 dso->filename);
+          return 1;
+        }
+      break;
+    case R_ARM_TLS_TPOFF32:
+      if (dso->ehdr.e_type == ET_EXEC)
+	error (0, 0, "%s: R_ARM_TLS_TPOFF32 relocs should not be present in "
+	       "prelinked ET_EXEC REL sections",
+	       dso->filename);
+      break;
     default:
       error (0, 0, "%s: Unknown arm relocation type %d", dso->filename,
 	     (int) GELF_R_TYPE (rel->r_info));
@@ -178,10 +197,10 @@ arm_prelink_rela (struct prelink_info *info, GElf_Rela *rela,
     {
     case R_ARM_GLOB_DAT:
     case R_ARM_JUMP_SLOT:
-      write_le32 (dso, rela->r_offset, value + rela->r_addend);
+      write_ne32 (dso, rela->r_offset, value + rela->r_addend);
       break;
     case R_ARM_ABS32:
-      write_le32 (dso, rela->r_offset, value + rela->r_addend);
+      write_ne32 (dso, rela->r_offset, value + rela->r_addend);
       break;
     case R_ARM_PC24:
       val = value + rela->r_addend - rela->r_offset;
@@ -192,8 +211,8 @@ arm_prelink_rela (struct prelink_info *info, GElf_Rela *rela,
 	  return 1;
 	}
       val &= 0xffffff;
-      write_le32 (dso, rela->r_offset,
-		  (read_ule32 (dso, rela->r_offset) & 0xff000000) | val);
+      write_ne32 (dso, rela->r_offset,
+		  (read_une32 (dso, rela->r_offset) & 0xff000000) | val);
       break;
     case R_ARM_COPY:
       if (dso->ehdr.e_type == ET_EXEC)
@@ -201,6 +220,24 @@ arm_prelink_rela (struct prelink_info *info, GElf_Rela *rela,
 	return 0;
       error (0, 0, "%s: R_ARM_COPY reloc in shared library?", dso->filename);
       return 1;
+    case R_ARM_TLS_DTPOFF32:
+      write_ne32 (dso, rela->r_offset, value + rela->r_addend);
+      break;
+    /* DTPMOD32 and TPOFF32 is impossible to predict in shared libraries
+       unless prelink sets the rules.  */
+    case R_ARM_TLS_DTPMOD32:
+      if (dso->ehdr.e_type == ET_EXEC)
+        {
+          error (0, 0, "%s: R_ARM_TLS_DTPMOD32 reloc in executable?",
+                 dso->filename);
+          return 1;
+        }
+      break;
+    case R_ARM_TLS_TPOFF32:
+      if (dso->ehdr.e_type == ET_EXEC && info->resolvetls)
+        write_ne32 (dso, rela->r_offset,
+                    value + rela->r_addend + info->resolvetls->offset);
+      break;
     default:
       error (0, 0, "%s: Unknown arm relocation type %d", dso->filename,
 	     (int) GELF_R_TYPE (rela->r_info));
@@ -218,7 +255,7 @@ arm_apply_conflict_rela (struct prelink_info *info, GElf_Rela *rela,
     case R_ARM_GLOB_DAT:
     case R_ARM_JUMP_SLOT:
     case R_ARM_ABS32:
-      buf_write_le32 (buf, rela->r_addend);
+      buf_write_ne32 (info->dso, buf, rela->r_addend);
       break;
     default:
       abort ();
@@ -240,14 +277,14 @@ arm_apply_rel (struct prelink_info *info, GElf_Rel *rel, char *buf)
       break;
     case R_ARM_GLOB_DAT:
     case R_ARM_JUMP_SLOT:
-      buf_write_le32 (buf, value);
+      buf_write_ne32 (info->dso, buf, value);
       break;
     case R_ARM_ABS32:
-      buf_write_le32 (buf, value + read_ule32 (info->dso, rel->r_offset));
+      buf_write_ne32 (info->dso, buf, value + read_une32 (info->dso, rel->r_offset));
       break;
     case R_ARM_PC24:
       val = value + rel->r_offset;
-      value = read_ule32 (info->dso, rel->r_offset) << 8;
+      value = read_une32 (info->dso, rel->r_offset) << 8;
       value = ((Elf32_Sword) value) >> 6;
       val += value;
       val >>= 2;
@@ -257,7 +294,7 @@ arm_apply_rel (struct prelink_info *info, GElf_Rel *rel, char *buf)
 	  return 1;
 	}
       val &= 0xffffff;
-      buf_write_le32 (buf, (buf_read_ule32 (buf) & 0xff000000) | val);
+      buf_write_ne32 (info->dso, buf, (buf_read_une32 (info->dso, buf) & 0xff000000) | val);
       break;
     case R_ARM_COPY:
       abort ();
@@ -285,7 +322,7 @@ arm_apply_rela (struct prelink_info *info, GElf_Rela *rela, char *buf)
     case R_ARM_GLOB_DAT:
     case R_ARM_JUMP_SLOT:
     case R_ARM_ABS32:
-      buf_write_le32 (buf, value + rela->r_addend);
+      buf_write_ne32 (info->dso, buf, value + rela->r_addend);
       break;
     case R_ARM_PC24:
       val = value + rela->r_addend - rela->r_offset;
@@ -296,7 +333,7 @@ arm_apply_rela (struct prelink_info *info, GElf_Rela *rela, char *buf)
 	  return 1;
 	}
       val &= 0xffffff;
-      buf_write_le32 (buf, (buf_read_ule32 (buf) & 0xff000000) | val);
+      buf_write_ne32 (info->dso, buf, (buf_read_une32 (info->dso, buf) & 0xff000000) | val);
       break;
     case R_ARM_COPY:
       abort ();
@@ -315,6 +352,7 @@ arm_prelink_conflict_rel (DSO *dso, struct prelink_info *info, GElf_Rel *rel,
 {
   GElf_Addr value;
   struct prelink_conflict *conflict;
+  struct prelink_tls *tls;
   GElf_Rela *ret;
 
   if (GELF_R_TYPE (rel->r_info) == R_ARM_RELATIVE
@@ -324,8 +362,34 @@ arm_prelink_conflict_rel (DSO *dso, struct prelink_info *info, GElf_Rel *rel,
   conflict = prelink_conflict (info, GELF_R_SYM (rel->r_info),
 			       GELF_R_TYPE (rel->r_info));
   if (conflict == NULL)
-    return 0;
-  value = conflict_lookup_value (conflict);
+    {
+      if (info->curtls == NULL)
+	return 0;
+
+      switch (GELF_R_TYPE (rel->r_info))
+	{
+	/* Even local DTPMOD and TPOFF relocs need conflicts.  */
+	case R_ARM_TLS_DTPMOD32:
+	case R_ARM_TLS_TPOFF32:
+	  break;
+
+	default:
+	  return 0;
+	}
+      value = 0;
+    }
+  else
+    {
+      /* DTPOFF32 wants to see only real conflicts, not lookups
+	 with reloc_class RTYPE_CLASS_TLS.  */
+      if (GELF_R_TYPE (rel->r_info) == R_ARM_TLS_DTPOFF32
+	  && conflict->lookup.tls == conflict->conflict.tls
+	  && conflict->lookupval == conflict->conflictval)
+        return 0;
+
+      value = conflict_lookup_value (conflict);
+    }
+
   ret = prelink_conflict_add_rela (info);
   if (ret == NULL)
     return 1;
@@ -345,6 +409,33 @@ arm_prelink_conflict_rel (DSO *dso, struct prelink_info *info, GElf_Rel *rel,
     case R_ARM_COPY:
       error (0, 0, "R_ARM_COPY should not be present in shared libraries");
       return 1;
+    case R_ARM_TLS_DTPMOD32:
+    case R_ARM_TLS_DTPOFF32:
+    case R_ARM_TLS_TPOFF32:
+      if (conflict != NULL
+	  && (conflict->reloc_class != RTYPE_CLASS_TLS
+	      || conflict->lookup.tls == NULL))
+	{
+	  error (0, 0, "%s: TLS reloc not resolving to STT_TLS symbol",
+		 dso->filename);
+	  return 1;
+	}
+      tls = conflict ? conflict->lookup.tls : info->curtls;
+      ret->r_info = GELF_R_INFO (0, R_ARM_ABS32);
+      switch (GELF_R_TYPE (rel->r_info))
+	{
+	case R_ARM_TLS_DTPMOD32:
+	  ret->r_addend = tls->modid;
+	  break;
+	case R_ARM_TLS_DTPOFF32:
+	  ret->r_addend = value;
+	  break;
+	case R_ARM_TLS_TPOFF32:
+	  ret->r_addend = (value + read_une32 (dso, rel->r_offset)
+			   + tls->offset);
+	  break;
+	}
+      break;
     default:
       error (0, 0, "%s: Unknown arm relocation type %d", dso->filename,
 	     (int) GELF_R_TYPE (rel->r_info));
@@ -359,6 +450,7 @@ arm_prelink_conflict_rela (DSO *dso, struct prelink_info *info,
 {
   GElf_Addr value;
   struct prelink_conflict *conflict;
+  struct prelink_tls *tls;
   GElf_Rela *ret;
   Elf32_Sword val;
 
@@ -368,9 +460,36 @@ arm_prelink_conflict_rela (DSO *dso, struct prelink_info *info,
     return 0;
   conflict = prelink_conflict (info, GELF_R_SYM (rela->r_info),
 			       GELF_R_TYPE (rela->r_info));
+
   if (conflict == NULL)
-    return 0;
-  value = conflict_lookup_value (conflict);
+    {
+      if (info->curtls == NULL)
+	return 0;
+
+      switch (GELF_R_TYPE (rela->r_info))
+	{
+	/* Even local DTPMOD and TPOFF relocs need conflicts.  */
+	case R_ARM_TLS_DTPMOD32:
+	case R_ARM_TLS_TPOFF32:
+	  break;
+
+	default:
+	  return 0;
+	}
+      value = 0;
+    }
+  else
+    {
+      /* DTPOFF32 wants to see only real conflicts, not lookups
+	 with reloc_class RTYPE_CLASS_TLS.  */
+      if (GELF_R_TYPE (rela->r_info) == R_ARM_TLS_DTPOFF32
+	  && conflict->lookup.tls == conflict->conflict.tls
+	  && conflict->lookupval == conflict->conflictval)
+        return 0;
+
+      value = conflict_lookup_value (conflict);
+    }
+
   ret = prelink_conflict_add_rela (info);
   if (ret == NULL)
     return 1;
@@ -391,13 +510,39 @@ arm_prelink_conflict_rela (DSO *dso, struct prelink_info *info,
 	  error (0, 0, "%s: R_ARM_PC24 overflow", dso->filename);
 	  return 1;
 	}
-      value = read_ule32 (dso, rela->r_offset) & 0xff000000;
+      value = read_une32 (dso, rela->r_offset) & 0xff000000;
       ret->r_addend = (Elf32_Sword) (value | (val & 0xffffff));
       ret->r_info = GELF_R_INFO (0, R_ARM_ABS32);
       break;
     case R_ARM_COPY:
       error (0, 0, "R_ARM_COPY should not be present in shared libraries");
       return 1;
+    case R_ARM_TLS_DTPMOD32:
+    case R_ARM_TLS_DTPOFF32:
+    case R_ARM_TLS_TPOFF32:
+      if (conflict != NULL
+	  && (conflict->reloc_class != RTYPE_CLASS_TLS
+	      || conflict->lookup.tls == NULL))
+	{
+	  error (0, 0, "%s: TLS reloc not resolving to STT_TLS symbol",
+		 dso->filename);
+	  return 1;
+	}
+      tls = conflict ? conflict->lookup.tls : info->curtls;
+      ret->r_info = GELF_R_INFO (0, R_ARM_ABS32);
+      switch (GELF_R_TYPE (rela->r_info))
+	{
+	case R_ARM_TLS_DTPMOD32:
+	  ret->r_addend = tls->modid;
+	  break;
+	case R_ARM_TLS_DTPOFF32:
+	  ret->r_addend = value;
+	  break;
+	case R_ARM_TLS_TPOFF32:
+	  ret->r_addend = value + rela->r_addend + tls->offset;
+	  break;
+	}
+      break;
     default:
       error (0, 0, "%s: Unknown arm relocation type %d", dso->filename,
 	     (int) GELF_R_TYPE (rela->r_info));
@@ -418,14 +563,17 @@ arm_rel_to_rela (DSO *dso, GElf_Rel *rel, GElf_Rela *rela)
       abort ();
     case R_ARM_RELATIVE:
     case R_ARM_ABS32:
-      rela->r_addend = (Elf32_Sword) read_ule32 (dso, rel->r_offset);
+    case R_ARM_TLS_TPOFF32:
+      rela->r_addend = (Elf32_Sword) read_une32 (dso, rel->r_offset);
       break;
     case R_ARM_PC24:
-      rela->r_addend = read_ule32 (dso, rel->r_offset) << 8;
+      rela->r_addend = read_une32 (dso, rel->r_offset) << 8;
       rela->r_addend = ((Elf32_Sword) rela->r_addend) >> 6;
       break;
     case R_ARM_COPY:
     case R_ARM_GLOB_DAT:
+    case R_ARM_TLS_DTPMOD32:
+    case R_ARM_TLS_DTPOFF32:
       rela->r_addend = 0;
       break;
     }
@@ -445,15 +593,18 @@ arm_rela_to_rel (DSO *dso, GElf_Rela *rela, GElf_Rel *rel)
       abort ();
     case R_ARM_RELATIVE:
     case R_ARM_ABS32:
-      write_le32 (dso, rela->r_offset, rela->r_addend);
+    case R_ARM_TLS_TPOFF32:
+      write_ne32 (dso, rela->r_offset, rela->r_addend);
       break;
     case R_ARM_PC24:
-      write_le32 (dso, rela->r_offset,
-		  (read_ule32 (dso, rela->r_offset) & 0xff000000)
+      write_ne32 (dso, rela->r_offset,
+		  (read_une32 (dso, rela->r_offset) & 0xff000000)
 		  | ((rela->r_addend >> 2) & 0xffffff));
       break;
     case R_ARM_GLOB_DAT:
-      write_le32 (dso, rela->r_offset, 0);
+    case R_ARM_TLS_DTPMOD32:
+    case R_ARM_TLS_DTPOFF32:
+      write_ne32 (dso, rela->r_offset, 0);
       break;
     }
   return 0;
@@ -479,7 +630,7 @@ arm_need_rel_to_rela (DSO *dso, int first, int last)
 	    switch (ELF32_R_TYPE (rel->r_info))
 	      {
 	      case R_ARM_ABS32:
-		val = read_ule32 (dso, rel->r_offset);
+		val = read_une32 (dso, rel->r_offset);
 		/* R_ARM_ABS32 with addend 0 can be converted
 		   to R_ARM_GLOB_DAT and we don't have to convert
 		   to RELA because of that.  */
@@ -488,6 +639,14 @@ arm_need_rel_to_rela (DSO *dso, int first, int last)
 		/* FALLTHROUGH */
 	      case R_ARM_PC24:
 		return 1;
+	      case R_ARM_TLS_TPOFF32:
+		/* In shared libraries TPOFF is changed always into
+		   conflicts, for executables we need to preserve
+		   original addend.  */
+		if (dso->ehdr.e_type == ET_EXEC)
+		  return 1;
+
+		break;
 	      }
 	}
     }
@@ -521,7 +680,7 @@ arm_arch_prelink (struct prelink_info *info)
       if (i == dso->ehdr.e_shnum)
 	return 0;
       data = dso->shdr[i].sh_addr;
-      write_le32 (dso, dso->info[DT_PLTGOT] + 4, data);
+      write_ne32 (dso, dso->info[DT_PLTGOT] + 4, data);
     }
 
   return 0;
@@ -550,9 +709,9 @@ arm_arch_undo_prelink (DSO *dso)
 
       if (i == dso->ehdr.e_shnum)
 	return 0;
-      data = read_ule32 (dso, dso->info[DT_PLTGOT] + 4);
+      data = read_une32 (dso, dso->info[DT_PLTGOT] + 4);
       if (data == dso->shdr[i].sh_addr)
-	write_le32 (dso, dso->info[DT_PLTGOT] + 4, 0);
+	write_ne32 (dso, dso->info[DT_PLTGOT] + 4, 0);
     }
 
   return 0;
@@ -580,17 +739,17 @@ arm_undo_prelink_rel (DSO *dso, GElf_Rel *rel, GElf_Addr reladdr)
 	}
       else
 	{
-	  Elf32_Addr data = read_ule32 (dso, dso->shdr[sec].sh_addr + 4);
+	  Elf32_Addr data = read_une32 (dso, dso->shdr[sec].sh_addr + 4);
 
 	  assert (rel->r_offset >= dso->shdr[sec].sh_addr + 12);
 	  assert (((rel->r_offset - dso->shdr[sec].sh_addr) & 3) == 0);
-	  write_le32 (dso, rel->r_offset, data);
+	  write_ne32 (dso, rel->r_offset, data);
 	}
       break;
     case R_ARM_GLOB_DAT:
       sec = addr_to_sec (dso, rel->r_offset);
 
-      write_le32 (dso, rel->r_offset, 0);
+      write_ne32 (dso, rel->r_offset, 0);
       if (sec != -1)
 	{
 	  if (strcmp (strptr (dso, dso->ehdr.e_shstrndx,
@@ -614,6 +773,12 @@ arm_undo_prelink_rel (DSO *dso, GElf_Rel *rel, GElf_Addr reladdr)
 	return 0;
       error (0, 0, "%s: R_ARM_COPY reloc in shared library?", dso->filename);
       return 1;
+    case R_ARM_TLS_DTPMOD32:
+    case R_ARM_TLS_DTPOFF32:
+      write_ne32 (dso, rel->r_offset, 0);
+      break;
+    case R_ARM_TLS_TPOFF32:
+      break;
     default:
       error (0, 0, "%s: Unknown arm relocation type %d", dso->filename,
 	     (int) GELF_R_TYPE (rel->r_info));
@@ -636,6 +801,10 @@ arm_reloc_class (int reloc_type)
     {
     case R_ARM_COPY: return RTYPE_CLASS_COPY;
     case R_ARM_JUMP_SLOT: return RTYPE_CLASS_PLT;
+    case R_ARM_TLS_DTPMOD32:
+    case R_ARM_TLS_DTPOFF32:
+    case R_ARM_TLS_TPOFF32:
+      return RTYPE_CLASS_TLS;
     default: return RTYPE_CLASS_VALID;
     }
 }
@@ -648,6 +817,7 @@ PL_ARCH = {
   .R_JMP_SLOT = R_ARM_JUMP_SLOT,
   .R_COPY = R_ARM_COPY,
   .R_RELATIVE = R_ARM_RELATIVE,
+  .rtype_class_valid = RTYPE_CLASS_VALID,
   .dynamic_linker = "/lib/ld-linux.so.2",
   .adjust_dyn = arm_adjust_dyn,
   .adjust_rel = arm_adjust_rel,
